﻿using System.Linq;
using SharpNeat.Core;
using SharpNeat.Decoders;
using SharpNeat.Genomes.Neat;
using SharpNeat.Network;
using SharpNeat.Phenomes;

namespace FaceDetection.Convolution
{
  public class ConvGenomeDecoder : IGenomeDecoder<NeatGenome, IBlackBox>
  {
    private readonly int windowSize;
    private readonly int span;
    private readonly int offset;
    private readonly uint[,] ids;

    public ConvGenomeDecoder(int windowSize, int span, int offset)
    {
      this.windowSize = windowSize;
      this.span = span;
      this.offset = offset;

      ids = new uint[windowSize, windowSize];
      uint id = 1;
      for (int y = 0; y < windowSize; ++y)
      {
        for (int x = 0; x < windowSize; ++x)
        {
          ids[y, x] = id++;
        }
      }
    }
    
    public IBlackBox Decode(NeatGenome genome)
    {
      return Decode(genome, genome.ActivationFnLibrary);
    }
    
    public IBlackBox Decode(NeatGenome genome, IActivationFunctionLibrary library)
    {
      var nodes = new NodeList(genome.NodeList.Count + windowSize * windowSize);
      var connections = new ConnectionList(genome.ConnectionList.Count + genome.InputNodeCount * span * span);
      uint idOffset = (uint)(windowSize * windowSize);

      nodes.Add(genome.NodeList[0]);
      for (int i = 0; i < windowSize * windowSize; ++i)
      {
        nodes.Add(new NetworkNode((uint)(i + 1), NodeType.Input, 0));
      }

      foreach (var node in genome.NodeList.Skip(1))
      {
        var type = node.NodeType != NodeType.Input ? node.NodeType : NodeType.Hidden;
        nodes.Add(new NetworkNode(node.Id + idOffset, type, 0));
      }

      int convIndex = 1;
      for (int hy = 0; hy + span <= windowSize; hy += offset)
      {
        for (int hx = 0; hx + span <= windowSize; hx += offset)
        {
          uint targetId = (uint)(windowSize * windowSize + convIndex);
          double[] aux = genome.NodeList[convIndex].AuxState;

          int auxIndex = 0;
          for (int iy = hy; iy < hy + span; ++iy)
          {
            for (int ix = hx; ix < hx + span; ++ix)
            {
              connections.Add(new NetworkConnection(ids[iy, ix], targetId, aux[auxIndex]));
              ++auxIndex;
            }
          }
          ++convIndex;
        }
      }

      for (int i = 0; i < genome.ConnectionList.Count; ++i)
      {
        var conn = genome.ConnectionList[i];
        uint sourceId = conn.SourceNodeId == 0 ? 0 : conn.SourceNodeId + idOffset;
        connections.Add(new NetworkConnection(sourceId, conn.TargetNodeId + idOffset, conn.Weight));
      }

      return FastAcyclicNetworkFactory.CreateFastAcyclicNetwork(new NetworkDefinition(windowSize * windowSize,
        genome.OutputNeuronCount, library, nodes, connections));
    }
  }
}
