﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FaceDetection.Neural;

namespace FaceDetection.Evolution
{
  public sealed class Population
  {
    private readonly Random random;
    private readonly GeneFactory factory;
    private readonly Clusterer clusterer;
    private readonly Func<Network, double> evaluate;

    private readonly double elitePart;
    private readonly double crossoverRate;
    private readonly double mutateRate;

    private readonly List<Genome> genomes;
    private readonly List<Cluster> clusters;

    public double AverageFitness { get; private set; }
    public double BestFitness { get; private set; }
    public Network BestNetwork { get; private set; }

    public Population(int size, double elitePart, double crossoverRate, double mutateRate,
      Random random, GeneFactory factory, Clusterer clusterer, Func<Network, double> evaluate)
    {
      this.random = random;
      this.factory = factory;
      this.clusterer = clusterer;
      this.evaluate = evaluate;

      this.elitePart = elitePart;
      this.crossoverRate = crossoverRate;
      this.mutateRate = mutateRate;

      genomes = new List<Genome>();
      clusters = new List<Cluster>();

      while (genomes.Count != size)
      {
        genomes.Add(new Genome(factory, random));
      }
      clusters = clusterer.InitClusters(genomes, random);
      EndEpoch();
    }

    public void AdvanceEpoch()
    {
      double totalFitness = clusters.Sum(c => c.Fitness);
      var offspringCount = new int[clusters.Count];
      var fraction = new Tuple<double, int>[clusters.Count];

      for (int i = 0; i < clusters.Count; ++i)
      {
        var realCount = clusters[i].Fitness / totalFitness * genomes.Count;
        offspringCount[i] = (int)realCount;
        fraction[i] = Tuple.Create(realCount - offspringCount[i], i);
      }

      int leftover = genomes.Count - offspringCount.Sum();
      foreach (int index in fraction.OrderByDescending(t => t.Item1).Select(t => t.Item2).Take(leftover))
      {
        ++offspringCount[index];
      }
    }

    public void EndEpoch()
    {
      clusterer.UpdateClusters(clusters, genomes);
      foreach (var cluster in clusters)
      {
        cluster.CalculateFitness();
      }

      BestFitness = Double.NegativeInfinity;
      foreach (var genome in genomes)
      {
        var network = genome.CreateNetwork();
        genome.Fitness = evaluate(network);
        if (genome.Fitness > BestFitness)
        {
          BestFitness = genome.Fitness;
          BestNetwork = network;
        }
      }
      AverageFitness = genomes.Average(g => g.Fitness);
    }
  }
}
