﻿namespace FaceDetection
{
  using System;
  using System.Collections.Generic;
  using System.Drawing;
  using System.Drawing.Drawing2D;
  using System.Drawing.Imaging;
  using System.IO;
  using System.Linq;
  using System.Runtime.InteropServices;
  
  public sealed class Loader
  {
    private const int CornerMask = 5;
    private const int BorderMask = 1;

    private int windowSize;
    private bool[] masked;

    private readonly List<byte[]> positives;
    private readonly List<byte[]> negatives;

    public Loader(bool loadNegatives = true)
    {
      windowSize = Common.WindowSize;
      masked = new bool[windowSize * windowSize];
      int index = 0;
      for (int y = 0; y < windowSize; ++y)
      {
        for (int x = 0; x < windowSize; ++x)
        {
          if (x < BorderMask || x >= windowSize - BorderMask || x + y < CornerMask ||
              x + windowSize - 1 - y < CornerMask || windowSize - 1 - x + y < CornerMask ||
              windowSize - 1 - x + windowSize - 1 - y < CornerMask)
          {
            masked[index] = true;
          }
          ++index;
        }
      }

      positives = LoadAllPositives();
      negatives = loadNegatives ? LoadAllNegatives() : new List<byte[]>();
    }

    public List<byte[]> GetPositives()
    {
      return positives;
    }

    public static int PositivesCount()
    {
      return Directory.EnumerateFiles(@"Images\Positives").Count();
    }

    public List<byte[]> GetNegatives()
    {
      return negatives;
    }

    private List<byte[]> LoadAllPositives()
    {
      Console.WriteLine("Loading positives...");
      return Directory.EnumerateFiles(@"Images\Positives")
        .Select(filename => LoadPositive(filename)).ToList();
    }

    private List<byte[]> LoadAllNegatives()
    {
      Console.WriteLine("Loading negatives...");
      var negatives = Directory.EnumerateFiles(@"Images\Negatives")
        .SelectMany(filename => LoadNegatives(filename, true));
      var positives = Directory.EnumerateFiles(@"Images\Positives")
        .SelectMany(filename => LoadNegatives(filename, false));

      return negatives.Concat(positives).ToList();
    }
    
    private byte[] LoadPositive(string path)
    {
      using (var original = Image.FromFile(path))
      using (var scaled = ScaleImage(original, windowSize, windowSize))
      {
        var bytes = GetBitmapBytes(scaled);
        return ProcessBytes(bytes, windowSize, 0, 0);
      }
    }

    public IEnumerable<Tuple<int, Rectangle, byte[]>> LoadWindows(string path)
    {
      using (var original = Image.FromFile(path))
      {
        var windows = DoLoadWindows(original, null);
        for (int i = 20; i <= 20; i += 10)
        {
          windows = windows.Concat(DoLoadWindows(original, i)).Concat(DoLoadWindows(original, -i));
        }
        foreach (var window in windows)
        {
          yield return window;
        }
      }
    }

    private IEnumerable<Tuple<int, Rectangle, byte[]>> DoLoadWindows(Image original, double? angle)
    {
      Console.Write("{0,4}° | ", angle.GetValueOrDefault(0));
      
      Image rotated;
      Matrix inverse;
      double dw, dh;

      if (angle != null)
      {
        Matrix rotate = new Matrix();
        rotate.Rotate((float)angle.Value);
        var points = new[] { new PointF(0, 0), new PointF(0, original.Height),
          new PointF(original.Width, 0), new PointF(original.Width, original.Height) };
        rotate.TransformPoints(points);
        dw = Math.Abs(points.Max(p => p.X) - points.Min(p => p.X)) / original.Width;
        dh = Math.Abs(points.Max(p => p.Y) - points.Min(p => p.Y)) / original.Height;

        rotated = new Bitmap((int)(original.Width * dw), (int)(original.Height * dh));
        using (var gfx = Graphics.FromImage(rotated))
        {
          gfx.InterpolationMode = InterpolationMode.HighQualityBicubic;
          gfx.CompositingQuality = CompositingQuality.HighQuality;
          gfx.SmoothingMode = SmoothingMode.HighQuality;

          gfx.TranslateTransform(-original.Width / 2, -original.Height / 2, MatrixOrder.Append);
          gfx.RotateTransform((float)angle.Value, MatrixOrder.Append);
          gfx.TranslateTransform(rotated.Width / 2, rotated.Height / 2, MatrixOrder.Append);
          gfx.DrawImage(new Bitmap(original), Point.Empty);
          inverse = gfx.Transform.Clone();
          inverse.Invert();
        }
      }
      else
      {
        rotated = original;
        inverse = null;
        dw = dh = 1;
      }

      double scale = 1;
      int level = 0;
      while (true)
      {
        int width = (int)(rotated.Width / scale);
        int height = (int)(rotated.Height / scale);
        if (width < windowSize || height < windowSize)
        {
          break;
        }

        using (var scaled = ScaleImage(rotated, width, height))
        {
          int fullWindow = (int)(windowSize * scale);
          int halfWindow = fullWindow / 2;
          var bytes = GetBitmapBytes(scaled);
          for (int y = 0; y + windowSize < height; ++y)
          {
            for (int x = 0; x + windowSize < width; ++x)
            {
              var point = new[] { new Point((int)(x * scale) + halfWindow, (int)(y * scale) + halfWindow) };
              if (inverse != null)
              {
                inverse.TransformPoints(point);
              }
              var rect = new Rectangle(point[0].X - halfWindow, point[0].Y - halfWindow, fullWindow, fullWindow);
              var screen = new Rectangle(0, 0, original.Width, original.Height);
              screen.Intersect(rect);
              if (screen.Width * screen.Height >= rect.Width * rect.Height / 2)
              {
                yield return Tuple.Create(level, rect, ProcessBytes(bytes, scaled.Width, x, y));
              }
            }
          }
        }
        scale *= Math.Sqrt(Math.Sqrt(2));
        ++level;
        Console.Write("*");
      }
      if (angle != null)
      {
        rotated.Dispose();
      }
      Console.WriteLine();
    }

    public IEnumerable<byte[]> LoadNegatives(string path, bool includeFull)
    {
      using (var original = Image.FromFile(path))
      {
        double scale = 1;
        while (true)
        {
          int width = (int)(original.Width / scale);
          int height = (int)(original.Height / scale);
          if (width < windowSize || height < windowSize)
          {
            break;
          }
          if (!includeFull && width < windowSize * 2 || height < windowSize * 2)
          {
            break;
          }

          using (var scaled = ScaleImage(original, width, height))
          {
            var bytes = GetBitmapBytes(scaled);
            for (int y = 0; y + windowSize < height; y += windowSize / 2)
            {
              for (int x = 0; x + windowSize < width; x += windowSize / 2)
              {
                yield return ProcessBytes(bytes, scaled.Width, x, y);
              }
            }
          }
          scale *= Math.Sqrt(2);
        }
      }
    }

    private Bitmap ScaleImage(Image bitmap, int width, int height)
    {
      var scaled = new Bitmap(width, height, PixelFormat.Format32bppRgb);
      using (var gfx = Graphics.FromImage(scaled))
      {
        gfx.InterpolationMode = InterpolationMode.HighQualityBicubic;
        gfx.CompositingQuality = CompositingQuality.HighQuality;
        gfx.SmoothingMode = SmoothingMode.HighQuality;

        var attrs = new ImageAttributes();
        attrs.SetWrapMode(WrapMode.TileFlipXY);
        gfx.DrawImage(bitmap, new Rectangle(0, 0, width, height),
          0, 0, bitmap.Width, bitmap.Height, GraphicsUnit.Pixel, attrs);
      }
      return scaled;
    }

    private byte[] GetBitmapBytes(Bitmap bitmap)
    {
      var data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppRgb);
      var color = new byte[bitmap.Width * bitmap.Height * 4];
      Marshal.Copy(data.Scan0, color, 0, color.Length);
      bitmap.UnlockBits(data);

      var gray = new byte[bitmap.Width * bitmap.Height];
      for (int i = 0; i < color.Length; i += 4)
      {
        gray[i / 4] = (byte)Math.Round(0.299 * color[i + 1] + 0.587 * color[i + 2] + 0.114 * color[i + 3]);
      }
      return gray;
    }

    private byte[] ProcessBytes(byte[] bytes, int width, int x, int y)
    {
      byte[] window = new byte[windowSize * windowSize];
      for (int i = 0; i < windowSize; ++i)
      {
        Array.Copy(bytes, (y + i) * width + x, window, i * windowSize, windowSize);
      }

      RemoveLighting(window);
      NormalizeHistogram(window);
      return window;
    }

    private void RemoveLighting(byte[] image)
    {
      double[,] m = new double[3, 4];

      int index = 0;
      for (int y = 0; y < windowSize; ++y)
      {
        for (int x = 0; x < windowSize; ++x)
        {
          int z = image[index];
          if (masked[index++])
          {
            continue;
          }

          m[0, 0] += x * x;
          m[0, 1] += x * y;
          m[0, 2] += x;
          m[1, 1] += y * y;
          m[1, 2] += y;

          m[0, 3] += x * z;
          m[1, 3] += y * z;
          m[2, 2] += 1;
          m[2, 3] += z;
        }
      }

      m[1, 0] = m[0, 1];
      m[2, 0] = m[0, 2];
      m[2, 1] = m[1, 2];

      for (int i = 0; i < 3; ++i)
      {
        for (int j = 3; j >= i; --j)
        {
          m[i, j] /= m[i, i];
        }

        for (int j = 0; j < 3; ++j)
        {
          if (j != i)
          {
            for (int k = 3; k >= i; --k)
            {
              m[j, k] -= m[j, i] * m[i, k];
            }
          }
        }
      }

      index = 0;
      for (int y = 0; y < windowSize; ++y)
      {
        for (int x = 0; x < windowSize; ++x)
        {
          int light = (int)Math.Round(x * m[0, 3] + y * m[1, 3]);
          image[index] = (byte)Math.Max(0, Math.Min(255, image[index] - light));
          ++index;
        }
      }
    }

    private void NormalizeHistogram(byte[] image)
    {
      var histogram = new int[256];
      int total = 0;
      for (int i = 0; i < image.Length; ++i)
      {
        if (!masked[i])
        {
          ++histogram[image[i]];
          ++total;
        }
      }

      var comulative = new int[256];
      comulative[0] = histogram[0];
      for (int i = 1; i < comulative.Length; ++i)
      {
        comulative[i] = comulative[i - 1] + histogram[i];
      }

      int min = comulative.Min();
      for (int i = 0; i < histogram.Length; ++i)
      {
        histogram[i] = (int)Math.Round((double)(comulative[i] - min) / ((double)total - min) * 255);
      }

      for (int i = 0; i < image.Length; ++i)
      {
        image[i] = (byte)histogram[image[i]];
      }
    }
  }
}
