﻿namespace FaceDetection
{
  using System;
  using System.IO;
  using System.Linq;
  using System.Threading;
  using System.Xml;
  using SharpNeat.Core;
  using SharpNeat.DistanceMetrics;
  using SharpNeat.EvolutionAlgorithms;
  using SharpNeat.EvolutionAlgorithms.ComplexityRegulation;
  using SharpNeat.Genomes.Neat;
  using SharpNeat.Phenomes;
  using SharpNeat.SpeciationStrategies;

  public static class Trainer
  {
    private static Loader loader;
    private static NeatEvolutionAlgorithm<NeatGenome> neat;
    private static Evaluator evaluator;
    private static string extraStats;
    private static StreamWriter log;

    private static void RunNeat(int alt, int iter, NeatGenomeFactory factory, IGenomeDecoder<NeatGenome, IBlackBox> decoder, double factor)
    {
      var basename = Path.Combine("Log", String.Format("{0:00}_{1:0000}", alt, iter));
      if (File.Exists(basename + ".xml"))
      {
        Console.WriteLine("skipped");
        return;
      }

      try
      {
        log = new StreamWriter(basename + ".txt");
      }
      catch
      {
        Console.WriteLine("skipped");
        return;
      }

      evaluator = new Evaluator(loader, factor);
      var parEvaluator = new ParallelGenomeListEvaluator<NeatGenome, IBlackBox>(decoder, evaluator);

      var metric = new ManhattanDistanceMetric(1.0, 0.0, 10.0);
      var speciation = new ParallelKMeansClusteringStrategy<NeatGenome>(metric);
      var complexity = new DefaultComplexityRegulationStrategy(ComplexityCeilingType.Relative, 10);

      neat = new NeatEvolutionAlgorithm<NeatGenome>(new NeatEvolutionAlgorithmParameters
      {
        SpecieCount = 25,
      }, speciation, complexity);

      neat.Initialize(parEvaluator, factory, 500);
      neat.UpdateScheme = new UpdateScheme(1U);
      neat.UpdateEvent += OnNeatUpdate;

      neat.StartContinue();

      while (neat.RunState == RunState.Running)
      {
        Thread.Sleep(TimeSpan.FromSeconds(1));
      }

      log.Close();

      using (var writer = XmlWriter.Create(basename + ".xml", new XmlWriterSettings { Indent = true }))
      {
        NeatGenomeXmlIO.WriteComplete(writer, neat.CurrentChampGenome, true);
      }
    }

    public static void Train()
    {
      var factories = Common.GetFactories();
      var decoders = Common.GetDecoders();
      loader = new Loader();
      Console.WriteLine();

      for (int i = 0; i < 30; ++i)
      {
        for (int j = 0; j < factories.Length; ++j)
        {
          Console.Write("{0,2} : {1,2} | ", i, j);
          RunNeat(j, i, factories[j](), decoders[j](), i < 8 ? 0.5 : 0.4);
        }
      }
    }

    private static void OnNeatUpdate(object sender, EventArgs e)
    {
      if (evaluator.StopConditionSatisfied)
      {
        return;
      }

      var stats = neat.Statistics;
      var auxFitness = neat.CurrentChampGenome.EvaluationInfo.AuxFitnessArr;

      if (stats._generation % 50 == 0)
      {
        extraStats = evaluator.FullyValidate((IBlackBox)neat.CurrentChampGenome.CachedPhenome);
      }

      log.WriteLine("{0}\t{1}\t{2:0.00000}\t{3:0.00000}\t{4:0.00000}\t{5}\t{6}\t{7}\t{8:0.00000}\t{9}\t{10:0.00000}",
        stats._generation, neat.ComplexityRegulationMode == ComplexityRegulationMode.Complexifying ? "C" : "S",
        stats._meanFitness, stats._meanComplexity,
        neat.CurrentChampGenome.EvaluationInfo.Fitness, neat.CurrentChampGenome.Complexity, extraStats,
        stats._minSpecieSize, neat.SpecieList.Average(s => s.GenomeList.Count), stats._maxSpecieSize,
        stats._meanSpecieChampFitness);

      log.Flush();

      if (stats._generation >= 1000)
      {
        Console.WriteLine("* |");
        evaluator.StopConditionSatisfied = true;
        return;
      }

      if (stats._generation % 50 == 0 && stats._generation > 0)
      {
        Console.Write("*");
        if (stats._generation % 250 == 0)
        {
          Console.Write(" ");
        }

        var boxes = neat.SpecieList
          .Select(s => s.GenomeList[0])
          .OrderByDescending(g => g.EvaluationInfo.Fitness)
          .Take(4)
          .Select(g => (IBlackBox)g.CachedPhenome)
          .ToList();

        evaluator.ExpandSets(boxes);
      }
    }
  }
}
