﻿using System;
using System.Collections.Generic;
using System.Linq;
using FaceDetection.Neural;

namespace FaceDetection.Evolution
{
  public sealed class Genome : IVector
  {
    private readonly List<NodeGene> nodes;
    private readonly List<LinkGene> links;

    public double Fitness { get; set; }
    public Cluster Cluster { get; set; }

    public Genome(GeneFactory factory, Random random)
    {
      const int weightRange = 2;

      nodes = factory.CreateInitialNodes();
      links = factory.CreateInitialLinks();
      foreach (var link in links)
      {
        link.Weight = random.NextDouble(weightRange);
      }
    }

    private Genome(List<NodeGene> nodes, List<LinkGene> links)
    {
      this.nodes = nodes;
      this.links = links;
    }

    public bool AddNode(GeneFactory factory, Random random)
    {
      const int maxTries = 5;
      for (int i = 0; i < maxTries; ++i)
      {
        var node = factory.CreateNode(links.GetRandom(random));
        if (nodes.InsertSorted(node))
        {
          return true;
        }
      }
      return false;
    }

    public bool AddLink(GeneFactory factory, Random random)
    {
      const int maxTries = 5;
      for (int i = 0; i < maxTries; ++i)
      {
        NodeGene from, to;
        do
        {
          from = nodes.GetRandom(random);
          to = nodes.GetRandom(random);
        }
        while (to.Depth >= from.Depth);
        if (links.InsertSorted(factory.CreateLink(from, to)))
        {
          return true;
        }
      }
      return false;
    }

    public void MutateWeights(double chance, double range, Random random)
    {
      foreach (var link in links)
      {
        if (random.NextDouble() < chance)
        {
          link.Weight += random.NextDouble(range);
        }
      }
    }

    public static Genome Crossover(Genome first, Genome second, Random random)
    {
      if (first.Fitness < second.Fitness)
      {
        return Crossover(second, first, random);
      }

      var links = new List<LinkGene>();
      int i = 0, j = 0;
      while (i < first.links.Count)
      {
        if (first.links[i].Id < second.links[j].Id)
        {
          links.Add(first.links[i]);
          ++i;
        }
        else if (first.links[i].Id > second.links[j].Id)
        {
          ++j;
        }
        else
        {
          if (random.Next(2) == 0)
          {
            links.Add(first.links[i]);
          }
          else
          {
            links.Add(second.links[j]);
          }
          ++i;
          ++j;
        }
      }

      return new Genome(first.nodes.ToList(), links);
    }

    public Network CreateNetwork()
    {
      return new Network(nodes, links);
    }

    int IVector.Count
    {
      get { return links.Count; }
    }

    int IVector.GetId(int index)
    {
      return links[index].Id;
    }
    
    double IVector.GetWeight(int index)
    {
      return links[index].Weight;
    }
  }
}
